# -*- coding: utf-8 -*-

"""common constants and functions"""

import logging

from collections.abc import Iterator
from dataclasses import dataclass
from datetime import datetime
from getpass import getuser
from os import DirEntry, sep
from pathlib import Path
from platform import node
from typing import Literal, Self


EMPTY = ""
LF = "\n"
PACKAGE = "uvartage"
POSIX = "posix"
UTF_8 = "utf-8"

ENV_DEFAULT_REPOSITORY = "UVARTAGE_DEFAULT_REPOSITORY"


@dataclass(frozen=True)
class DirectoryEntry:
    """New Directory entry class with constructors from a DirEntry or a Path"""

    name: str
    entry_type: Literal["-", "d", "l"]
    size: int
    last_changed: str
    target: str | None = None

    @property
    def hidden(self) -> bool:
        """returns true if the file or dir is hidden"""
        return self.name.startswith(".")

    def sort_name(self) -> str:
        """return name"""
        return self.name

    def sort_last_changed(self) -> str:
        """return last chnaged time"""
        return self.last_changed

    def __str__(self) -> str:
        """return name"""
        return "  ".join(
            (
                f"{self.entry_type:.<10s}",
                f"{self.size:>10d}",
                self.last_changed,
                f"{self.name} → {self.target}" if self.entry_type == "l" else self.name,
            )
        )

    def display(self, long_format: bool = False) -> str:
        """self display"""
        if long_format:
            return str(self)
        #
        return self.name

    @classmethod
    def from_entry(cls, entry: DirEntry | Path, full_path: bool = False) -> Self:
        """Store attributes"""
        if isinstance(entry, Path):
            stat = entry.lstat()
        elif isinstance(entry, DirEntry):
            stat = entry.stat(follow_symlinks=False)
        else:
            raise TypeError("Must be a DirEntry or Path instance")
        #
        last_changed = datetime.fromtimestamp(stat.st_mtime_ns / 1e9).strftime(
            "%Y-%m-%d %H:%M:%S"
        )
        name = str(entry) if full_path else entry.name
        if entry.is_dir():
            return cls(f"{name}{sep}", "d", stat.st_size, last_changed)
        #
        if entry.is_symlink():
            return cls(
                name,
                "l",
                stat.st_size,
                last_changed,
                target=str(Path(entry).readlink()),
            )
        #
        return cls(name, "-", stat.st_size, last_changed)


def complete_paths(text: str, line: str, begidx: int, unused_endidx: int) -> list[str]:
    """Completion for paths on the command line"""
    results: list[str] = []
    if text.startswith("-"):
        # no completion for options
        return results
    #
    if text == "..":
        results.append("../")
        return results
    #
    pattern = f"{text}*"
    basepath = Path()
    if begidx > 0:
        preceding_argument = line[:begidx].split()[-1]
        if preceding_argument[-1] == sep:
            basepath = Path(preceding_argument)
        #
    #
    for found_path in basepath.glob(pattern):
        found_name = found_path.name
        if found_path.is_dir():
            results.append(f"{found_name}{sep}")
        else:
            results.append(found_name)
        #
    #
    return results


def enforce_str(source) -> str:
    """Return source if it is a str instance, or raise a TypeError"""
    if isinstance(source, str):
        return source
    #
    raise TypeError(f"Expected {source!r} to be of type str")


def get_prompt() -> str:
    """Return a prompt including user@host along with the current workdir"""
    return f"«{PACKAGE}» [{getuser()}@{node()} {Path.cwd().name}] "


def get_existing_paths(*files_or_dirs: str) -> list[Path]:
    """Return a list of existing paths from file_or_dirs
    after username and wildcard expansion
    """
    existing_paths: list[Path] = []
    error_messages: list[str] = []
    for single_pattern in files_or_dirs:
        if any((wildcard in single_pattern for wildcard in "?*")):
            if single_pattern.startswith(sep):
                basepath = Path(sep)
                single_pattern = single_pattern[1:]
            else:
                basepath = Path()
            #
            matched_paths = list(basepath.glob(single_pattern))
            if not matched_paths:
                error_messages.append(f"{single_pattern} not found")
            #
            existing_paths.extend(matched_paths)
        else:
            current_path = Path(single_pattern)
            if current_path.exists():
                existing_paths.append(current_path)
            else:
                error_messages.append(f"{single_pattern} does not exist")
            #
        #
    #
    if error_messages and not existing_paths:
        raise ValueError(LF.join(error_messages))
    #
    return existing_paths


def iter_paths_list(
    *files_or_dirs: str,
    long_format: bool = False,
    show_all: bool = False,
    sort_by_time: bool = False,
    sort_reverse: bool = False,
) -> Iterator[str]:
    """Iterate over paths"""
    existing_paths: list[Path] = get_existing_paths(*files_or_dirs)
    if not existing_paths:
        existing_paths.append(Path("."))
    #
    sort_key = DirectoryEntry.sort_name
    if sort_by_time:
        sort_key = DirectoryEntry.sort_last_changed
        sort_reverse = not sort_reverse
    #
    for candidate_path in existing_paths:
        output_entries: list[DirectoryEntry] = []
        if candidate_path.is_dir():
            entries = [
                DirectoryEntry.from_entry(dir_entry)
                for dir_entry in candidate_path.glob("*")
            ]
            for entry in sorted(entries, key=sort_key, reverse=sort_reverse):
                if show_all or not entry.hidden:
                    output_entries.append(entry)
                #
            #
            yield f"--- Contents of {candidate_path}:"
            for entry in output_entries:
                yield entry.display(long_format=long_format)
            #
            yield EMPTY
        else:
            try:
                yield DirectoryEntry.from_entry(candidate_path, full_path=True).display(
                    long_format=long_format
                )
            except OSError as error:
                logging.error(str(error))
            #
        #
    #


def log_multiline(message: str, level: int = logging.INFO) -> None:
    """Log message, using multiple log calls if necessary for multimple lines"""
    for line in message.splitlines():
        logging.log(level, line)
    #



# -*- coding: utf-8 -*-

"""uvartage command line interface"""

import argparse
import getpass
import importlib.metadata
import logging
import os
import pathlib
import sys
import tomllib

from .backends import SupportedBackendType
from .commons import PACKAGE, UTF_8, ENV_DEFAULT_REPOSITORY
from .loop import Loop


def get_project_file_contents(project_file_name: str) -> str:
    """Return the contents of a project file two directories up"""
    project_directory_path = pathlib.Path(__file__).resolve().parent.parent.parent
    project_file_path = project_directory_path / project_file_name
    return project_file_path.read_text(encoding=UTF_8)


def get_metadata_version(metadata_file_name: str = "pyproject.toml") -> str:
    """get version information from metadata_file up_dirs directories above"""
    metadata = tomllib.loads(get_project_file_contents(metadata_file_name))
    version = metadata["project"]["version"]
    return f"{version} (read directly from {metadata_file_name})"


def get_arguments(*args: str, test_context: bool = False) -> argparse.Namespace:
    """Parse commandline arguments"""
    try:
        version = importlib.metadata.version(PACKAGE)
    except importlib.metadata.PackageNotFoundError:
        version = get_metadata_version()
    #
    parser = argparse.ArgumentParser(
        prog=PACKAGE,
        description="Wrapper for uv with artifact storage in airgapped environments",
    )
    parser.set_defaults(loglevel=logging.WARNING, user=getpass.getuser())
    parser.add_argument(
        "--version", action="version", version=version, help="print version and exit"
    )
    parser.add_argument(
        "-v ",
        "--verbose",
        action="store_const",
        const=logging.INFO,
        dest="loglevel",
        help="show more messages",
    )
    parser.add_argument(
        "--backend",
        type=SupportedBackendType,
        # see https://stackoverflow.com/a/46385352>
        choices=list(SupportedBackendType),
        default=SupportedBackendType.ARTIFACTORY,
        help="the artifact storage backend type"
        " (default and currently the only supported backend: %(default)s)",
    )
    parser.add_argument(
        "--ca-file",
        type=pathlib.Path,
        help="a CA cert bundle file to be provided via SSL_CERT_FILE.",
    )
    parser.add_argument(
        "--user",
        help="username for the artifact storage backend if the hostname is"
        " not explicitly specified as USER@HOSTNAME; default is %(default)r.",
    )
    parser.add_argument(
        "hostname",
        metavar="[USER@]HOSTNAME",
        help="the artifact storage hostname, or user and hostname combined by '@'.",
    )
    parser.add_argument(
        "repositories",
        nargs=argparse.REMAINDER,
        help="the package repositories (default first)."
        " If not at least one repository name is provided, the value of the"
        f" environment variable {ENV_DEFAULT_REPOSITORY} will be used.",
    )
    if args or test_context:
        arguments = parser.parse_args(list(args))
    else:
        arguments = parser.parse_args()
    #
    logging.basicConfig(format="| %(message)s", level=arguments.loglevel)
    return arguments


def main() -> int:
    """Execute the loop and return a zero returncode"""
    arguments = get_arguments()
    if not arguments.repositories:
        try:
            default_repository = os.environ[ENV_DEFAULT_REPOSITORY]
        except KeyError:
            logging.critical(
                "If no repository names are provided on the command line,"
                " a default repository name has to be specified via"
                " environment variable %r.",
                ENV_DEFAULT_REPOSITORY,
            )
            return 1
        #
        arguments.repositories.append(default_repository)
    #
    repl = Loop(
        ca_file=arguments.ca_file,
        hostname_argument=arguments.hostname,
        repositories=arguments.repositories,
        default_username=arguments.user,
    )
    # Do not exit on KeyboardInterrupt
    repeat = True
    intro = "Welcome to the uv wrapper shell. Type help or ? to list commands.\n"
    while repeat:
        try:
            repl.cmdloop(intro=intro)
        except ValueError as error:
            logging.critical(str(error))
            return 1
        #
        except KeyboardInterrupt:
            # silent REPL restart
            print("^C")
            intro = ""
            continue
        #
        repeat = False
    #
    return 0


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-

"""uvartage: wrapper for uv with artifact storage in airgapped environments"""

import logging

from enum import Enum
from getpass import getpass
from os import environ
from pathlib import Path
from subprocess import run, CompletedProcess
from sys import orig_argv
from urllib.parse import urlunsplit, quote

from .commons import EMPTY, enforce_str


_SSL_CERT_FILE = "SSL_CERT_FILE"
_MASK = "[MASKED]"

PIP = "pip"
UV = "uv"
UVX = "uvx"
_EXTRA = "extra"

PYTHON = orig_argv[0]


def get_index_url(base_url: str, repository_name: str) -> str:
    """Index URL for the provided repository name"""
    return f"{base_url}{repository_name}/simple"


class SupportedBackendType(str, Enum):
    """Supported Backend types"""

    ARTIFACTORY = "artifactory"

    def __str__(self):
        """required for proper argparse display,
        compare <https://stackoverflow.com/a/46385352>
        """
        return self.value


# pylint: disable=too-many-instance-attributes


class BackendBase:
    """Backend base class"""

    api_path = "/"
    scheme = "https"

    def __init__(
        self,
        hostname_argument: str,
        default_username: str,
        *repositories: str,
        require_login: bool = True,
        ca_file: Path | None = None,
    ) -> None:
        """Store the URL parts and the username"""
        self.__base_environment = dict(environ)
        if isinstance(ca_file, Path):
            ca_full_path = ca_file.resolve()
            if not ca_full_path.is_file():
                raise ValueError(f"{ca_full_path} does not exist or is not a file")
            #
            self.__base_environment.update(SSL_CERT_FILE=str(ca_full_path))
        elif _SSL_CERT_FILE not in self.__base_environment:
            logging.warning(
                "Neither the environment variable %r has been set,"
                " nor a CA file explicitly through --ca-file.",
                _SSL_CERT_FILE,
            )
            logging.warning(
                "You might encounter problems if using a non-standard"
                " (i.e. organization internal) certificate authority."
            )
        #
        try:
            self.__username, self.__hostname = hostname_argument.split("@")
        except ValueError:
            self.__username = default_username
            self.__hostname = hostname_argument
        #
        self.__api_base_url = enforce_str(
            urlunsplit((self.scheme, self.__hostname, self.api_path, EMPTY, EMPTY))
        )
        if require_login:
            self.__password = getpass(
                f"Please enter the password for {self.__username}"
                f" on {self.__hostname} (input is hidden): "
            )
            if not self.__password:
                raise ValueError("Stopping due to empty password input")
            #
        else:
            self.__username = EMPTY
            self.__password = EMPTY
        #
        self.__environment_additions: dict[str, dict[str, str]] = {_EXTRA: {}}
        self.__masked_overrides: dict[str, str] = {}
        self.__deleted_keys: set[str] = set()
        self.repositories = tuple(repositories)

    @property
    def hostname(self) -> str:
        """hostname property"""
        return self.__hostname

    @property
    def username(self) -> str:
        """Username property"""
        return self.__username

    @property
    def api_base_url(self) -> str:
        """API base URL"""
        return self.__api_base_url

    @property
    def deleted_keys(self) -> set[str]:
        """all deleted keys"""
        return set(self.__deleted_keys)

    @property
    def all_available_keys(self) -> set[str]:
        """all keys available as environment variables, except deleted keys"""
        return self.keys() - self.__deleted_keys

    def _get_environment(self, *include_additions: str) -> dict[str, str]:
        """Return the enwironment including the named additions"""
        environment: dict[str, str] = dict(self.__base_environment)
        for single_addition in [
            item for item in include_additions if item != _EXTRA
        ] + [_EXTRA]:
            environment |= self._get_additions(single_addition)
        #
        # Subtract deletions
        return {
            key: value
            for key, value in environment.items()
            if key in set(environment) - self.__deleted_keys
        }

    def get_masked_environment(self, *include_additions: str) -> dict[str, str]:
        """Return the enwironment including the named additions,
        values masked if necessary
        """
        return {
            key: self.__masked_overrides.get(key, value)
            for key, value in self._get_environment(*include_additions).items()
        }

    def execute(
        self,
        command: str,
        *args: str,
        check: bool = True,
        capture_output: bool = False,
        shell: bool = False,
    ) -> CompletedProcess:
        """Execute a command with the matching environment"""
        include_additions: list[str] = []
        if command in (PYTHON, PIP):
            include_additions.append(PIP)
        #
        if command in (UV, UVX):
            include_additions.append(UV)
        #
        environment = self._get_environment(*include_additions)
        if shell:
            ignoring = f" (ignoring {args})" if args else ""
            logging.warning("Executing with shell=True: %r%s", command, ignoring)
            return run(
                command,
                env=environment,
                check=check,
                capture_output=capture_output,
                shell=True,
            )
        #
        full_command = [command, *args]
        logging.info("Executing command: %r", full_command)
        return run(
            full_command,
            env=environment,
            check=check,
            capture_output=capture_output,
            shell=False,
        )

    def keys(self) -> set[str]:
        """Return all keys as a set, including deleted ones"""
        found_keys = set(self.__base_environment)
        for addition_key in (PIP, UV, _EXTRA):
            found_keys |= set(self._get_additions(addition_key))
        #
        return found_keys

    def _make_pip_environment(self) -> dict[str, str]:
        """Return the pip environment with PIP_INDEX_URL, and
        (if necessary) PIP_CERT and/or PIP_EXTRA_INDEX_URL
        """
        additions: dict[str, str] = {}
        ssl_cert_file = self.__base_environment.get(_SSL_CERT_FILE)
        if ssl_cert_file:
            additions.setdefault("PIP_CERT", ssl_cert_file)
        #
        if self.__username:
            pip_api_base_url = enforce_str(
                urlunsplit(
                    (
                        self.scheme,
                        f"{quote(self.__username, safe=EMPTY)}:"
                        f"{quote(self.__password, safe=EMPTY)}@"
                        f"{self.__hostname}",
                        self.api_path,
                        EMPTY,
                        EMPTY,
                    )
                )
            )
            masked_pip_api_base_url = enforce_str(
                urlunsplit(
                    (
                        self.scheme,
                        f"{quote(self.__username, safe=EMPTY)}:{_MASK}@{self.__hostname}",
                        self.api_path,
                        EMPTY,
                        EMPTY,
                    )
                )
            )
        else:
            pip_api_base_url = self.__api_base_url
            masked_pip_api_base_url = EMPTY
        #
        additions["PIP_INDEX_URL"] = get_index_url(
            pip_api_base_url, self.repositories[0]
        )
        extra_urls = [
            get_index_url(pip_api_base_url, extra_repo)
            for extra_repo in self.repositories[1:]
        ]
        masked_extra_urls = []
        if masked_pip_api_base_url:
            self.__masked_overrides["PIP_INDEX_URL"] = get_index_url(
                masked_pip_api_base_url, self.repositories[0]
            )
            masked_extra_urls = [
                get_index_url(masked_pip_api_base_url, extra_repo)
                for extra_repo in self.repositories[1:]
            ]
        #
        if extra_urls:
            additions["PIP_EXTRA_INDEX_URL"] = " ".join(extra_urls)
            if masked_extra_urls:
                self.__masked_overrides["PIP_EXTRA_INDEX_URL"] = " ".join(
                    masked_extra_urls
                )
            #
        #
        return additions

    def _make_uv_environment(self) -> dict[str, str]:
        """Return the uv environment with UV_DEFAULT_INDEX,
        UV_INDEX containing all extra indexes if any are defined
        (ie. if the number of repositiories is greater than 1),
        and UV_INDEX_{KEY}_USERNAME and UV_INDEX_{KEY}_PASSWORD for each index key
        (ie. "primary" for the default index with sequence number 0,
        and "extraN" for each extra index where N is its sequence number)
        """
        additions: dict[str, str] = {}
        extra_indexes = []
        for index_number, index_repo in enumerate(self.repositories):
            index_key = f"extra{index_number}" if index_number else "primary"
            index_cred_prefix = f"UV_INDEX_{index_key.upper()}"
            current_index = (
                f"{index_key}={get_index_url(self.__api_base_url, index_repo)}"
            )
            if index_number:
                extra_indexes.append(current_index)
            else:
                additions["UV_DEFAULT_INDEX"] = current_index
            #
            if self.__username:
                additions.update(
                    {
                        f"{index_cred_prefix}_USERNAME": self.__username,
                        f"{index_cred_prefix}_PASSWORD": self.__password,
                    }
                )
                self.__masked_overrides[f"{index_cred_prefix}_PASSWORD"] = _MASK
            #
        #
        if extra_indexes:
            additions.update(UV_INDEX=" ".join(extra_indexes))
        #
        return additions

    def _get_additions(self, flavor: str) -> dict[str, str]:
        """Return or set the environment additions"""
        try:
            return self.__environment_additions[flavor]
        except KeyError:
            logging.debug("Setting environment additions for %s ...", flavor)
        #
        if flavor == UV:
            additions_creator_function = self._make_uv_environment
        elif flavor == PIP:
            additions_creator_function = self._make_pip_environment
        else:
            raise ValueError(f"unsupported flavor {flavor!r}")
        #
        return self.__environment_additions.setdefault(
            flavor, additions_creator_function()
        )

    def delete_entry(self, key: str) -> None:
        """Mark key as deleted"""
        self.__deleted_keys.add(key)

    def recover_entry(self, key: str) -> None:
        """Recover and rea previously deleted entry"""
        if key in self.all_available_keys:
            logging.info("%s was still available", key)
            return
        #
        try:
            self.__deleted_keys.remove(key)
        except KeyError:
            logging.warning("%s not found")
        #

    def set_extra_entry(self, key: str, value: str) -> None:
        """Add or update an entry in the "extra" environment additions"""
        if not value:
            raise ValueError(f"Empty value, will not add or update {key}")
        #
        self.__environment_additions[_EXTRA][key] = value

    def set_masked_extra_entry(self, key: str) -> None:
        """Add or update a masked entry in the "extra" environment additions"""
        secret_value = getpass(f"Please enter the value for {key} (input is hidden): ")
        self.set_extra_entry(key, secret_value)
        self.__masked_overrides[key] = _MASK


class Artifactory(BackendBase):
    """Artifactory backend"""

    api_path = "/artifactory/api/pypi/"


def get_backend(
    backend_type: SupportedBackendType,
    hostname_argument: str,
    default_username: str,
    *repositories: str,
    ca_file: Path | None = None,
) -> BackendBase:
    """Return the backend for backend_type ans hostname_argument"""
    match backend_type:
        case SupportedBackendType.ARTIFACTORY:
            return Artifactory(
                hostname_argument, default_username, *repositories, ca_file=ca_file
            )
        #
    #
    raise ValueError(f"Unsupported backend {backend_type!r}")



# -*- coding: utf-8 -*-

"""uvartage cmd loop implementation"""

import logging

from argparse import ArgumentParser, Namespace, REMAINDER
from cmd import Cmd
from dataclasses import dataclass
from fnmatch import filter as fnmatch_filter
from os import chdir, getcwd
from pathlib import Path
from re import IGNORECASE, compile as re_compile
from shlex import split
from subprocess import CalledProcessError
from sys import stderr

from .backends import PIP, PYTHON, UV, BackendBase, SupportedBackendType, get_backend
from .commons import (
    EMPTY,
    UTF_8,
    complete_paths,
    get_prompt,
    iter_paths_list,
    log_multiline,
)


ID_PATTERN = "[a-z][_a-z0-9]*"

PRX_VARIABLE_ASSIGNMENT = re_compile(
    rf"\A(?P<key>{ID_PATTERN})=(?P<value>.*)\Z",
    IGNORECASE,
)
PRX_VARIABLE_NAME = re_compile(rf"\A{ID_PATTERN}\Z", IGNORECASE)

_PYTHONPATH = "PYTHONPATH"


def confirmed(prompt: str) -> bool:
    """Ask for confirmation"""
    prompt = f"{prompt} (y/n)? "
    while True:
        answer = input(prompt).strip().lower()
        if not answer:
            continue
        #
        if "yes".startswith(answer):
            return True
        #
        if "no".startswith(answer):
            return False
        #
        prompt = "Please answer yes or no: "
    #


class OptionError(Exception):
    """Raised if the OptionParser would exit"""


@dataclass
class Availability:
    """Programs availability"""

    backend: BackendBase
    git: str | None = None
    pip: str | None = None
    uv: str | None = None

    def __post_init__(self):
        """Find out program versions"""
        if self.git is None:
            self.git = self.get_stripped_output("git", "--version")
        #
        if self.pip is None:
            self.pip = self.get_stripped_output(PYTHON, "-m", "pip", "--version")
        #
        if self.pip is None:
            self.get_stripped_output(PYTHON, "-m", "ensurepip")
            self.pip = self.get_stripped_output(PYTHON, "-m", "pip", "--version")
        #
        if self.uv is None:
            self.uv = self.get_stripped_output("uv", "self", "version")
        #
        if self.uv is None and confirmed(
            "uv does not seem to be installed. Install now using pip"
        ):
            self.get_stripped_output(PYTHON, "-m", "pip", "install", "uv")
            self.uv = self.get_stripped_output("uv", "self", "version")
        #

    def get_stripped_output(self, *command: str) -> str | None:
        """Return the stripped command output, or None in case of an error"""
        try:
            captured_stdout = self.backend.execute(
                *command, capture_output=True, check=True
            ).stdout
        except CalledProcessError:
            return None
        #
        if isinstance(captured_stdout, bytes):
            return captured_stdout.decode(UTF_8).strip()
        #
        if isinstance(captured_stdout, str):
            return captured_stdout.strip()
        #
        raise TypeError(f"Unexpected return value: {captured_stdout!r}")

    def show_versions(self) -> None:
        """Show versions"""
        python_version = self.get_stripped_output(PYTHON, "--version")
        logging.info("%s (executable is %s)", python_version, PYTHON)
        logging.info(" * %s", self.pip or "no pip")
        logging.info(" * %s", self.uv or "no uv")
        logging.info(self.git or "no git")


class DefusedArgumentParser(ArgumentParser):
    """Defused argument parser not exiting the whoöe loop on wrong options"""

    msg_secret_value = "Secret values must be entered through hidden input"

    def exit(self, status=0, message=None):
        """Do not exit the program, raise an OptionError instead"""
        if message:
            stderr.write(message)
        #
        raise OptionError(status)

    @classmethod
    def get_list_arguments(cls, arg: str) -> Namespace:
        """Return parsed arguments for the list command"""
        parser = cls(prog="list", description="Print directory contents")
        parser.add_argument(
            "-l",
            "--long",
            action="store_true",
            help="long format (including size and last changed date)",
        )
        parser.add_argument(
            "-a",
            "--all",
            action="store_true",
            help="all files (by default, hidden files are not listed)",
        )
        parser.add_argument(
            "-r", "--reverse", action="store_true", help="reverse sort order"
        )
        parser.add_argument(
            "-t", "--time", action="store_true", help="sort by time instead of by name"
        )
        parser.add_argument(
            "files_or_dirs", nargs=REMAINDER, help="the file(s) or dir(s) to list"
        )
        return parser.parse_args(split(arg))

    @classmethod
    def get_env_arguments(cls, arg: str) -> tuple[tuple[str, ...], list[str]]:
        """Return parsed arguments for the env command"""
        parser = cls(prog="env", description="Print environment variables")
        parser.add_argument(
            "--include",
            nargs="*",
            choices=(PIP, UV),
            help="include additional variables (possible values: %(choices)s)",
        )
        parser.add_argument(
            "name_or_pattern",
            nargs="*",
            default=["*"],
            help="the variable name or pattern to show (default: %(default)s)",
        )
        arguments = parser.parse_args(split(arg))
        includes = tuple(arguments.include or [])
        return includes, arguments.name_or_pattern

    @classmethod
    def get_set_arguments(cls, arg: str) -> tuple[bool, str, str]:
        """Return parsed arguments for the env command"""
        parser = cls(prog="set", description="Set an environment variable")
        parser.add_argument(
            "-s",
            "--secret",
            action="store_true",
            help="enter the value through password input and mask it in putput",
        )
        parser.add_argument(
            "key",
            help="the variable name",
        )
        parser.add_argument(
            "value",
            nargs=REMAINDER,
            help="the variable value ",
        )
        arguments = parser.parse_args(split(arg))
        key = arguments.key
        explicit_value = " ".join(arguments.value)
        if arguments.secret and explicit_value:
            # Not allowed,
            # eg. argument is "-s KEY explicit value"
            parser.error(cls.msg_secret_value)
        #
        assignment_match = PRX_VARIABLE_ASSIGNMENT.match(key)
        if assignment_match:
            members = assignment_match.groupdict()
            key = members["key"]
            new_value = members["value"]
            if arguments.secret and (new_value or explicit_value):
                # Not allowed,
                # eg. argument is "-s KEY=value"
                parser.error(cls.msg_secret_value)
            #
            if explicit_value and not new_value:
                # eg. argument is 'KEY= explicit value'
                #     → KEY set to "explicit value"
                new_value = explicit_value
            elif new_value and explicit_value:
                # eg. argument is "KEY=new and explicit value"
                #     → KEY set to "new and explicit value"
                new_value = f"{new_value} {explicit_value}"
            #
            return arguments.secret, key, new_value
        #
        if not PRX_VARIABLE_NAME.match(key):
            # Not allowed,
            # eg. argument is "24 hours"
            parser.error(f"Invalid variable name {key}")
        #
        if arguments.secret:
            # eg. argument is "-s KEY"
            return True, key, EMPTY
        #
        if not explicit_value:
            explicit_value = input(f"Please enter a value for {key}: ")
        #
        # eg. argument is "KEY explicit value"
        #     → KEY is set to "explicit value"
        # eg. argument is "KEY" and "input value" is entered a the prompt →
        #     → KEY is set to "input value"
        return False, key, explicit_value


class Loop(Cmd):
    """Command loop"""

    def __init__(
        self,
        ca_file: Path | None,
        hostname_argument: str,
        repositories: list[str],
        default_username: str,
    ) -> None:
        """Initialize with the provided arguments"""
        if not repositories:
            raise ValueError("At least one repository is required")
        #
        self._backend = get_backend(
            SupportedBackendType.ARTIFACTORY,
            hostname_argument,
            default_username,
            *repositories,
            ca_file=ca_file,
        )
        self.availability = Availability(backend=self._backend)
        self.availability.show_versions()
        self.prompt = get_prompt()
        super().__init__()

    def do_cd(self, arg) -> None:
        """Change directory"""
        if arg:
            chdir(arg)
        else:
            try:
                chdir(Path.home())
            except RuntimeError as error:
                logging.error(str(error))
            #
        #
        self.prompt = get_prompt()

    def do_python(self, arg) -> None:
        """Run python (see output of python -V for the exact version)"""
        self._backend.execute(PYTHON, *split(arg), check=False)

    def do_spp(self, arg) -> None:
        """Shortcut for set PYTHONPATH=<arg>, eg. spp src → set PYTHONPATH=src"""
        try:
            self.do_set(f"{_PYTHONPATH}={split(arg)[0]}")
        except IndexError:
            logging.warning("spp requires an argument")
        #

    def complete_recover(
        self, text: str, unused_line: str, unused_begidx: int, unused_endidx: int
    ) -> list[str]:
        """Completion for the recover command"""
        if not text:
            return []
        #
        return sorted(key for key in self._backend.deleted_keys if key.startswith(text))

    def do_recover(self, arg) -> None:
        """Recover one or more previously unset environment variables"""
        for single_arg in split(arg):
            found_keys = sorted(fnmatch_filter(self._backend.deleted_keys, single_arg))
            for key in found_keys:
                logging.info("Recovering %s", key)
                self._backend.recover_entry(key)
            #
        #

    def do_unset(self, arg) -> None:
        """Unset (ie. reversibly delete) one or more environment variables"""
        all_keys = self._backend.keys()
        for single_arg in split(arg):
            found_keys = sorted(fnmatch_filter(all_keys, single_arg))
            for key in found_keys:
                logging.info("Unsetting %s", key)
                self._backend.delete_entry(key)
            #
        #

    def do_set(self, arg) -> None:
        """Set an environment variable"""
        try:
            secret, key, value = DefusedArgumentParser.get_set_arguments(arg)
        except OptionError:
            return
        #
        display = f"{key} → {value!r}"
        try:
            if secret:
                self._backend.set_masked_extra_entry(key)
                display = f"hidden value for {key}"
            else:
                self._backend.set_extra_entry(key, value)
            #
        except ValueError as error:
            logging.warning(str(error))
        else:
            logging.info("set %s", display)
        #

    def do_env(self, arg) -> None:
        """Print the environment variables"""
        try:
            include_additions, names_or_patterns = (
                DefusedArgumentParser.get_env_arguments(arg)
            )
        except OptionError:
            return
        #
        environment = self._backend.get_masked_environment(*include_additions)
        show_keys: set[str] = set()
        for single_pattern in names_or_patterns:
            show_keys.update(fnmatch_filter(environment, single_pattern))
        #
        if not show_keys:
            logging.warning(" – no matches –")
        #
        for key in sorted(show_keys):
            print(f"{key}={environment[key]!r}")
        #

    def complete_sh(
        self, text: str, line: str, begidx: int, unused_endidx: int
    ) -> list[str]:
        """Completion for the sh command"""
        return complete_paths(text, line, begidx, unused_endidx)

    def do_sh(self, arg) -> None:
        """Run an arbitrary command through the shell"""
        logging.warning("Running command with shell=True: %r", arg)
        self._backend.execute(arg, check=False, shell=True)

    def do_git(self, arg) -> None:
        """Run git with the provided arguments"""
        self._backend.execute("git", *split(arg), check=False)

    def complete_list(
        self, text: str, line: str, begidx: int, unused_endidx: int
    ) -> list[str]:
        """Completion for the list command"""
        return complete_paths(text, line, begidx, unused_endidx)

    def do_list(self, arg) -> None:
        """Print directory contents (emulation)"""
        try:
            args = DefusedArgumentParser.get_list_arguments(arg)
        except OptionError:
            return
        #
        try:
            for line in iter_paths_list(
                *args.files_or_dirs,
                long_format=args.long,
                show_all=args.all,
                sort_reverse=args.reverse,
                sort_by_time=args.time,
            ):
                print(line)
            #
        except ValueError as error:
            log_multiline(str(error), level=logging.ERROR)
            return
        #

    def do_pwd(self, unused_arg) -> None:
        """Print working directory"""
        if unused_arg:
            logging.warning("Ignored argument(s) %r", unused_arg)
        #
        print(getcwd())

    def do_pip(self, arg) -> None:
        """Run pip with the provided arguments"""
        self._backend.execute(PYTHON, "-m", "pip", *split(arg), check=False)

    def do_uv(self, arg) -> None:
        """Run uv with the provided arguments"""
        self._backend.execute("uv", *split(arg), check=False)

    def do_uvx(self, arg) -> None:
        """Run uvx with the provided arguments"""
        self._backend.execute("uvx", *split(arg), check=False)

    # pylint: disable=invalid-name ; required to support EOF character
    def do_EOF(self, unused_arg) -> bool:
        """Exit the REPL by EOF (eg. Ctrl-D on Unix)"""
        print()
        if unused_arg:
            logging.warning("Ignored argument(s) %r", unused_arg)
        #
        logging.info("bye")
        return True

    def emptyline(self) -> bool:
        """do nothing on empty input"""
        return False
